#!/usr/bin/env python
# coding: utf-8
#%%
"""
Simple client code for the ESP32 in Python
Copyright 2021 Benedict Diederich, released under LGPL 3.0 or later
"""

#import numpy as np



from .galvo import Galvo
from .config import config
from .logger import Logger
from .mserial import Serial
from .ledmatrix import LedMatrix
from .motor import Motor
from .state import State
from .laser import Laser
from .wifi import Wifi
from .camera import Camera


try:
    from imswitch.imcommon.model import initLogger
    IS_IMSWITCH = True
except:
    print("No imswitch available")
    IS_IMSWITCH = False


class UC2Client(object):
    # headers = {'ESP32-version': '*'}
    headers={"Content-Type":"application/json"}
    getmessage = ""
    is_connected = False

    is_wifi = False
    is_serial = False

    BAUDRATE = 115200
    
    def __init__(self, host=None, port=31950, serialport=None, identity=None, baudrate=BAUDRATE):
        '''
        This client connects to the UC2-REST microcontroller that can be found here
        https://github.com/openUC2/UC2-REST

        generally speaking you send/receive JSON documents that will cause an:
        1. action => "/XXX_act"
        2. getting => "/XXX_get"
        3. setting => "/XXX_set"

        you can send commands through wifi/http or usb/serial
        '''
    
        if IS_IMSWITCH:
            self.logger = initLogger(self, tryInheritParent=True)
        else:
            self.logger = Logger()            

        # initialize communication channel (# connect to wifi or usb)
        if serialport is not None:
            # use USB connection
            self.serial = Serial(serialport, baudrate, parent=self, identity=identity)
            self.is_serial = True
            self.is_connected = True
            
        elif host is not None:
            # use client in wireless mode
            self.is_wifi = True
            self.host = host
            self.port = port

            # check if host is up
            self.logger.debug(f"Connecting to microscope {self.host}:{self.port}")
            self.is_connected = self.isConnected()
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")

        #FIXME
        #self.set_state(debug=False)

        # initialize galvos
        #self.galvo1 = Galvo(channel=1) FIXME
        #self.galvo2 = Galvo(channel=2) FIXME
        
        # initialize config
        self.config = config(self)
        self.pinConfig = self.config.loadConfigDevice()
        
        # initialize LED matrix
        try: NLeds = self.pinConfig["ledArrNum"]
        except: NLeds=64
        self.led = LedMatrix(self, NLeds=NLeds)
        
        # initilize motor
        self.motor = Motor(self)
        
        # initialize laser
        self.state = State(self)
        
        # initialize galvo
        self.galvo1 = Galvo(self, 1)
        
        # initialize laser
        self.laser = Laser(self)
        
        # initialize wifi
        self.wifi = Wifi(self)
        
        # initialize camera
        self.camera = Camera(self)

    def post_json(self, path, payload, timeout=1):
        if self.is_wifi:
            # FIXME: this is not working
            url = f"http://{self.host}:{self.port}{path}"
            r = requests.post(url, json=payload, headers=self.headers)
            return r.json()
        elif self.is_serial:
            return self.serial.post_json(path, payload, timeout=timeout)
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")
            return None
        
    def get_json(self, path, timeout=1):
        if self.is_wifi:
            # FIXME: this is not working
            url = f"http://{self.host}:{self.port}{path}"
            r = requests.get(url, headers=self.headers)
            return r.json()
        elif self.is_serial:
            self.serial.get_json(self, path)
            return self.serial.read_json()
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")
            return None













import numpy as np
import time 

class Motor(object):
    
    microsteppingfactor_filter=16 # run more smoothly
    filter_pos_1 = 1000*microsteppingfactor_filter # GFP
    filter_pos_2 = 0*microsteppingfactor_filter # AF647/SIR
    filter_pos_3 = 500*microsteppingfactor_filter
    filter_pos_LED = filter_pos_1 # GFP / Brightfield
    filter_pos_init = -1250*microsteppingfactor_filter
    filter_speed = microsteppingfactor_filter * 500
    filter_position_now = 0


    def __init__(self, parent=None):
        self.steps_last_0 = 0
        self.steps_last_1 = 0
        self.steps_last_2 = 0
        self.steps_last_3 = 0
        
        self.backlashX = 0
        self.backlashY = 0
        self.backlashZ = 0
        self.backlashT = 0

        self.stepSizeX = 1
        self.stepSizeY = 1
        self.stepSizeZ = 1
        self.stepSizeT = 1

        self.maxStepX = np.inf
        self.minStepX = -np.inf
        self.maxStepY = np.inf
        self.minStepY = -np.inf
        self.maxStepZ = np.inf
        self.minStepZ = -np.inf
        self.maxStepT = np.inf
        self.minStepT = -np.inf
        self._parent = parent
        
        self.minPosX = -np.inf
        self.minPosY = -np.inf
        self.minPosZ = -np.inf
        self.minPosT = -np.inf
        self.maxPosX = np.inf
        self.maxPosY = np.inf
        self.maxPosZ = np.inf
        self.maxPosT = np.inf
        self.stepSizeX =  1
        self.stepSizeY =  1
        self.stepSizeZ =  1
        self.stepSizeT =  1
        
        
        
    '''################################################################################################################################################
    HIGH-LEVEL Functions that rely on basic REST-API functions
    ################################################################################################################################################'''
    def setup_motor(self, axis, minPos, maxPos, stepSize, backlash):
        if axis == "X":
            self.minPosX = minPos
            self.maxPosX = maxPos
            self.stepSizeX = stepSize
            self.backlashX = backlash
        elif axis == "Y":
            self.minPosY = minPos
            self.maxPosY = maxPos
            self.stepSizeY = stepSize
            self.backlashY = backlash
        elif axis == "Z":
            self.minPosZ = minPos
            self.maxPosZ = maxPos
            self.stepSizeZ = stepSize
            self.backlashZ = backlash
        elif axis == "T":
            self.minPosT = minPos
            self.maxPosT = maxPos
            self.stepSizeT = stepSize
            self.backlashT = backlash            
            
    def home_x(self):
        r = self.home(axis="X")
        return r
    
    def home_y(self):
        r = self.home(axis="Y")
        return r
    
    def home_z(self):
        r = self.home(axis="Z")
        return r
    
    def home_xyz(self):
        r = self.home(axis="XYZ")
        return r

    def home(self, axis="X"):
        path = "/motor_act",
        payload = {
            "task": path,
            "axis": axis
        }
        r = self._parent.post_json(path, payload)
        return r
        

    def move_x(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=np.inf):
        r = self.move_stepper(steps=(steps,0,0,0), speed=(speed,0,0,0), timeout=timeout, backlash=(self.backlashX,0,0,0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled)
        return r

    def move_y(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=np.inf):
        r = self.move_stepper(steps=(0,steps,0,0), speed=(0,speed,0,0), timeout=timeout, backlash=(0,self.backlashY,0,0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled)
        return r

    def move_z(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=np.inf):
        r = self.move_stepper(steps=(0,0,steps,0), speed=(0,0,speed,0), timeout=timeout, backlash=(0,0,self.backlashZ,0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled)
        return r

    def move_t(self, steps=100, speed=1000, is_blocking=False, is_absolute=False, is_enabled=True, timeout=np.inf):
        r = self.move_stepper(steps=(0,0,0,steps), speed=(0,0,0,speed), timeout=timeout, backlash=(0,0,self.backlashZ,0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled)
        return r
    
    def move_xyz(self, steps=(0,0,0), speed=(1000,1000,1000), is_blocking=False, is_absolute=False, is_enabled=True, timeout=np.inf):
        if len(speed)!= 3:
            speed = (speed,speed,speed)

        r = self.move_xyzt(steps=(steps[0],steps[1],steps[2],0), speed=(speed[0],speed[1],speed[2],0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)
        return r

    def move_xy(self, steps=(0,0), speed=(1000,1000), is_blocking=False, is_absolute=False, is_enabled=True, timeout=np.inf):
        if len(speed)!= 2:
            speed = (speed,speed)

        r = self.move_xyzt(steps=(steps[0],steps[1],0,0), speed=(speed[0],speed[1],0,0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)
        return r

    def move_xyzt(self, steps=(0,0,0,0), speed=(1000,1000,1000,1000), is_blocking=False, is_absolute=False, is_enabled=True, timeout=np.inf):
        if type(speed)==int:
            speed = (speed,speed,speed,speed)
        if type(steps)==int:
            steps = (steps,steps,steps,steps)
        
        r = self.move_stepper(steps=steps, speed=speed, backlash=(self.backlashX,self.backlashY,self.backlashZ,self.backlashT), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout)
        return r

    def init_filter(self, nSteps, speed=250, filter_axis=-1, is_blocking = True, is_enabled=False):
        self.move_filter(steps=nSteps, speed=speed, filter_axis=filter_axis, is_blocking=is_blocking, is_enabled = is_enabled)
        self.is_filter_init = True
        self.filter_position_now = 0

    def switch_filter(self, filter_pos=0, filter_axis=-1, timeout=20, is_filter_init=None, speed=None, is_enabled=False, is_blocking=True):

        # switch off all lasers first!
        self.set_laser(1, 0)
        self.set_laser(2, 0)
        self.set_laser(3, 0)

        if speed is None:
            speed = self.filter_speed

        if is_filter_init is not None:
            self.is_filter_init = is_filter_init
        if not self.is_filter_init:
            self.init_filter(nSteps=self.filter_pos_init, speed=speed, filter_axis=filter_axis, is_blocking = True)

        # measured in steps from zero position
        steps = filter_pos - self.filter_position_now
        self.filter_position_now = filter_pos

        self.move_filter(steps=steps, speed=speed, filter_axis=filter_axis, is_blocking=is_blocking, timeout=timeout, is_enabled=is_enabled)


    def move_filter(self, steps=100, speed=200, filter_axis=-1, timeout=10, is_enabled=False, is_blocking=False):
        steps_xyzt = np.zeros(4)
        steps_xyzt[filter_axis] = steps
        r = self.move_stepper(steps=steps_xyzt, speed=speed, timeout=timeout, is_enabled=is_enabled, is_blocking=is_blocking)
        return r




    def move_forever(self, speed=(0,0,0), is_stop=False):
        path = "/motor_act"
        payload = {
            "task":path,
            "speed0": np.int(speed[0]), # TODO: need a fourth axis?
            "speed1": np.int(speed[0]),
            "speed2": np.int(speed[1]),
            "speed3": np.int(speed[2]),
            "isforever":1,
            "isaccel":1,
            "isstop": np.int(is_stop)
        }
        # Make sure PS controller is treated correclty
        #if self.isControllerMode():
        #    self.setControllerMode(isController=False)
        #    PSwasActive = True
        #else:
        #    PSwasActive = False

        r = self._parent.post_json(path, payload, timeout=0)

        #if PSwasActive:
        #    self.setControllerMode(isController=True)

        return r

    def move_stepper(self, steps=(0,0,0,0), speed=(1000,1000,1000,1000), is_absolute=False, timeout=1, backlash=(0,0,0,0), is_blocking=True, is_enabled=True):
        '''
        This tells the motor to run at a given speed for a specific number of steps; Multiple motors can run simultaneously
        '''
        if type(speed)!=list and type(speed)!=tuple  :
            speed = (speed,speed,speed,speed)

        path = "/motor_act"

        # detect change in directiongit config --global user.name "Your Name"
        if np.sign(self.steps_last_0) != np.sign(steps[0]):
            # we want to overshoot a bit
            steps_0 = steps[0] + (np.sign(steps[0])*backlash[0])
        else: steps_0 = steps[0]
        if np.sign(self.steps_last_1) != np.sign(steps[1]):
            # we want to overshoot a bit
            steps_1 =  steps[1] + (np.sign(steps[1])*backlash[1])
        else: steps_1 = steps[1]
        if np.sign(self.steps_last_2) != np.sign(steps[2]):
            # we want to overshoot a bit
            steps_2 =  steps[2] + (np.sign(steps[2])*backlash[2])
        else: steps_2 = steps[2]
        if np.sign(self.steps_last_3) != np.sign(steps[3]):
            # we want to overshoot a bit
            steps_3 =  steps[3] + (np.sign(steps[3])*backlash[3])
        else: steps_3 = steps[3]
        
        # get current position
        pos_0 = self.get_position(3)
        pos_1 = self.get_position(0)
        pos_2 = self.get_position(1)
        pos_3 = self.get_position(2)

        # convert to physical units
        steps_0 = steps_0*self.stepSizeX
        steps_1 = steps_1*self.stepSizeY
        steps_2 = steps_2*self.stepSizeZ
        steps_3 = steps_3*self.stepSizeT
        
        # check if within limits
        if pos_0+steps_0 > self.maxPosX or pos_0+steps_0 < self.minPosX:
            steps_0=0
        if pos_1+steps_1 > self.maxPosY or pos_1+steps_1 < self.minPosY:
            steps_1=0
        if pos_2+steps_2 > self.maxPosZ or pos_2+steps_2 < self.minPosZ:
            steps_2 = 0
        if pos_3+steps_3 > self.maxPosT or pos_3+steps_3 < self.minPosT:
            steps_3 = 0
        
        payload = {
            "task":"/motor_act",
            "pos1": np.int(steps_0),
            "pos2": np.int(steps_1),
            "pos3": np.int(steps_2),
            "pos0": np.int(steps_3),
            "isblock": int(is_blocking),
            "isabs": int(is_absolute),
            "speed0": np.int(speed[3]),
            "speed1": np.int(speed[0]),
            "speed2": np.int(speed[1]),
            "speed3": np.int(speed[2]),
            "isen": np.int(is_enabled)
        }
        
        # safe steps to track direction for backlash compensatio
        self.steps_last_0 = steps_0
        self.steps_last_1 = steps_1
        self.steps_last_2 = steps_2
        self.steps_last_3 = steps_3

        # drive motor
        r = self._parent.post_json(path, payload, timeout=1)

        #if PSwasActive:
        #    self.setControllerMode(isController=True)
        # wait until job has been done
        time0=time.time()
        if is_blocking:
            while self._parent.state.isBusy():
                time.sleep(0.1)
                if time.time()-time0>timeout:
                    break

        return r


    def set_motor_maxSpeed(self, axis=0, maxSpeed=10000):
        path = "/motor_set",
        payload = {
            "task": path,
            "axis": axis,
            "maxspeed": maxSpeed
        }
        r = self._parent.post_json(path, payload)
        return r

    def set_motor_currentPosition(self, axis=0, currentPosition=10000):
        path = "/motor_set",
        payload = {
            "task": path,
            "axis": axis,
            "currentposition": currentPosition
        }
        r = self._parent.post_json(path, payload)
        return r

    def set_motor_acceleration(self, axis=0, acceleration=10000):
        path = "/motor_set",
        payload = {
            "task": path,
            "axis": axis,
            "acceleration": acceleration
        }
        r = self._parent.post_json(path, payload)
        return r


    def set_motor_enable(self, is_enable=1):
        path = "/motor_set",
        payload = {
            "task": path,
            "isen": is_enable
        }
        r = self._parent.post_json(path, payload)
        return r

    def set_direction(self, axis=1, sign=1, timeout=1):
        path = "/motor_set"

        payload = {
            "task":path,
            "axis": axis,
            "sign": sign
        }

        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def get_position(self, axis=1, timeout=1):
        path = "/motor_get"

        payload = {
            "task":path,
            "axis": axis
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try: _position = r["position"]
        except: _position = 0
        return _position

    def set_position(self, axis=1, position=0, timeout=1):
        path = "/motor_set"
        if axis=="X": axis=1
        if axis=="Y": axis=2
        if axis=="Z": axis=3

        payload = {
            "task":path,
            "axis":axis,
            "currentposition": position
        }
        r = self._parent.post_json(path, payload, timeout=timeout)

        return r




import json


class config(object):
    def __init__(self, parent, configFilePath=None): 
        self._parent = parent
        self.configFilePath = configFilePath
        
        if self.configFilePath is not None:
            self.configFilePath = "./config.json"
            #TODO: read file and load
            
        self.configFile = self.loadDefaultConfig()
        
    
    '''
    Set Configuration
    '''
    
    def setMotorConfig(self, motXstp, motXdir, motYstp, motYdir, motZstp, motZdir, motAstp, motAdir, motEnable):
        self.configFile["motXstp"] = motXstp
        self.configFile["motXdir"] = motXdir
        self.configFile["motYstp"] = motYstp
        self.configFile["motYdir"] = motYdir
        self.configFile["motZstp"] = motZstp
        self.configFile["motZdir"] = motZdir
        self.configFile["motAstp"] = motAstp
        self.configFile["motAdir"] = motAdir
        self.configFile["motEnable"] = motEnable
        
    def setDigitalPin(self, pin, value):
        self.configFile["digitalPin" + str(pin)] = value
        
    def setAnalogPin(self, pin, value):
        self.configFile["analogPin" + str(pin)] = value
        
    def setLEDArrayConfig(self, ledArrPin, ledArrNum):
        configFile = {}
        configFile["ledArrPin"] = ledArrPin
        configFile["ledArrNum"] = ledArrNum
        self.setConfigDevice(configFile, timeout=1)
        
    def setLaserPinConfig(self, pin1, pin2, pin3):
        self.configFile["laserPin1"] = pin1
        self.configFile["laserPin2"] = pin2
        self.configFile["laserPin3"] = pin3
        
    def setDACFakeConfig(self, pin1, pin2):
        self.configFile["dacFake1"] = pin1
        self.configFile["dacFake2"] = pin2
        
    def setWIFIconfig(self, ssid, PW):
        self.configFile["ssid"] = ssid
        self.configFile["PW"] = PW
        
    def setIdentifier(self, identifier):
        self.configFile["identifier"] = identifier
        
    '''
    get Configuration
    '''
    def getMotorPinConfig(self):
        return [self.configFile["motXstp"], 
                self.configFile["motXdir"], 
                self.configFile["motYstp"], 
                self.configFile["motYdir"], 
                self.configFile["motZstp"], 
                self.configFile["motZdir"], 
                self.configFile["motAstp"], 
                self.configFile["motAdir"], 
                self.configFile["motEnable"]]
    
    def getLaserPinConfig(self):
        return [self.configFile["laserPin1"], self.configFile["laserPin2"], self.configFile["laserPin3"]]
    
    def getLEDArrayConfig(self):
        return [self.configFile["ledArrPin"], self.configFile["ledArrNum"]]
    
    def getWiFiConfig(self):
        return [self.configFile["ssid"], self.configFile["PW"]]
    
    def getIdentifier(self):
        return self.configFile["identifier"]
    
    def getAnalogPinConfig(self):
        return [self.configFile["analogPin1"], self.configFile["analogPin2"], self.configFile["analogPin3"]]
    
    def getDigitalPin(self, pin):
        return self.configFile["digitalPin" + str(pin)]
    
    '''
    default Configuration
    '''     
        
    def loadDefaultConfig(self):
        self.defaultConfig = {
            "motXstp": 0,
            "motXdir": 0,
            "motYstp": 0,
            "motYdir": 0,
            "motZstp": 0,
            "motZdir": 0,
            "motAstp": 0,
            "motAdir": 0,
            "motEnable": 0,
            "ledArrPin": 0,
            "ledArrNum": 64,
            "digitalPin1":0,
            "digitalPin2":0,
            "analogPin1":0,
            "analogPin2":0,
            "analogPin3":0,
            "laserPin1":0,
            "laserPin2":0,
            "laserPin3":0,
            "dacFake1":0,
            "dacFake2":0,
            "identifier": "TEST",
            "ssid": "ssid",
            "PW": "PW"
        }
        return self.defaultConfig
    
    def setDefaultConfig(self, configDict=None, configFile=None):
        if configDict is not None:
            self.defaultConfig = configDict
        if configFile is not None:
            self.defaultConfig = self.json2dict(configFile)
    
    # importing the module


    def json2dict(self, jsonFile):
        with open(jsonFile) as f:
            data = json.load(f)
        return data
        
    
    '''
    send configuration to device
    '''
    
    '''################################################################################################################################################
    Set Configurations
    ################################################################################################################################################'''

    def loadConfigDevice(self, timeout=1):
        path = '/config_get'
        payload = {
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        
        self.setDefaultConfig(r)
        
        if type(r) != dict:
            r = self.loadDefaultConfig()
        return r

    def setConfigDevice(self, config, timeout=1):
        path = '/config_set'
        if type(config)==dict:
            payload = config
        else: 
            return None
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r


    


class digital(object):

    def __init__(self, parent):
        self._parent = parent
        self._logger = parent._logger
    
    def sendTrigger(self, triggerId=0):
        path = '/digital_act'

        payload = {
            "task": path,
            "digitalid": triggerId,
            "digitalval": -1,
        }

        r = self.post_json(path, payload)
        return r



import numpy as np

class LedMatrix(object):
    def __init__(self, parent, NLeds=64):
        #TOOD: This is for the LED matrix only!
        self.NLeds = NLeds
        #assuming we have a square grid!
        self.Nx = self.Ny = int(np.sqrt(NLeds))
        
        # we assume the pattern is binary (e.g. 0 or 1)
        self.ledpattern = np.zeros((self.Nx, self.Ny, 3))
        
        # this is a istance of the _parent32Client class
        self._parent = parent
        self.timeout = 1
        self.intensity = (255,255,255)
        
    
    '''
    ##############################################################################################################################
    LED ARRAY
    ##############################################################################################################################
    '''
    def send_LEDMatrix_array(self, led_pattern, timeout=1):
        '''
        Send an LED array pattern e.g. an RGB Matrix: led_pattern=np.zeros((3,8,8))
        '''
        path = '/ledarr_act'
        # Make sure LED strip is filled with matrix information
        if len(led_pattern.shape)<3:
            led_pattern = np.reshape(led_pattern, (led_pattern.shape[0], int(np.sqrt(led_pattern.shape[1])), int(np.sqrt(led_pattern.shape[1]))))
        led_pattern[:,1::2, :] = led_pattern[:,1::2, ::-1]
        payload = {
            "red": led_pattern[:,:,0].flatten().tolist(),
            "green": led_pattern[:,:,1].flatten().tolist(),
            "blue": led_pattern[:,:,2].flatten().tolist(),
            "arraySize": led_pattern.shape[1],
            "LEDArrMode": "array"
        }
        self._parent.logger.debug("Setting LED Pattern (array) ")
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def send_LEDMatrix_full(self, intensity = (255,255,255),timeout=1):
        '''
        set all LEDs with te same RGB value: intensity=(255,255,255)
        '''
        path = '/ledarr_act'
        payload = {
            "task":path,
            "red": int(intensity[0]),
            "green": int(intensity[1]),
            "blue": int(intensity[2]),
            "LEDArrMode": "full"
        }
        self._parent.logger.debug("Setting LED Pattern (full): "+ str(intensity))
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def send_LEDMatrix_special(self, pattern="left", intensity = (255,255,255),timeout=1):
        '''
        set all LEDs inside a certain pattern (e.g. left half) with the same RGB value: intensity=(255,255,255), rest 0
        '''
        path = '/ledarr_act'
        payload = {
            "red": int(intensity[0]),
            "green": int(intensity[1]),
            "blue": int(intensity[2]),
            "LEDArrMode": pattern
        }
        self._parent.logger.debug("Setting LED Pattern (full): "+ str(intensity))
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def send_LEDMatrix_single(self, indexled=0, intensity=(255,255,255), timeout=1):
        '''
        update only a single LED with a colour:  indexled=0, intensity=(255,255,255)
        '''
        path = '/ledarr_act'
        payload = {
            "red": int(intensity[0]),
            "green": int(intensity[1]),
            "blue": int(intensity[2]),
            "indexled": int(indexled),
            "LEDArrMode": "single"
        }
        self._parent.logger.debug("Setting LED PAttern: "+str(indexled)+" - "+str(intensity))
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def send_LEDMatrix_multi(self, indexled=(0), intensity=((255,255,255)), timeout=1):
        '''
        update a list of individual LEDs with a colour:  led_pattern=(1,2,6,11), intensity=((255,255,255),(125,122,1), ..)
        '''
        path = '/ledarr_act'
        payload = {
            "red": intensity[0],
            "green": intensity[1],
            "blue": intensity[2],
            "indexled": indexled,
            "LEDArrMode": "multi"
        }
        self._parent.logger.debug("Setting LED PAttern: "+str(indexled)+" - "+str(intensity))
        r = self._parent.post_json(path, payload, timeout=timeout)


    def get_LEDMatrix(self, timeout=1):
        '''
        get information about pinnumber and number of leds
        '''
        # TOOD: Not implemented yet
        path = "/ledarr_get"
        payload = {
            "task":path
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        return r

    def setSingle(self, indexled, state):
        # Update the intensity of a single LED in a cartesian grid        
        ix = indexled//self.Nx
        iy = indexled%self.Nx
        self.ledpattern[ix, iy] = state # either [0, 1]
        # forward backward enumaration
        if ix%2 != 0:
           indexled = (ix*self.Nx)+(self.Ny-iy-1)
        self.send_LEDMatrix_single(indexled=indexled, intensity=np.array(state)*np.array(self.intensity), timeout=self.timeout)

        return self.ledpattern
    
    def setAll(self, state, intensity=None):
        # fast addressing
        # turns on all LEDs at a certain intensity
        if intensity is not None:
            self.intensity = intensity
        intensity2display = np.array(self.intensity)*np.array(state)
        self.send_LEDMatrix_full(intensity = intensity2display, timeout=self.timeout)
        self.ledpattern = state*np.ones((self.Nx, self.Ny, 3))
        return self.ledpattern
    
    def setIntensity(self, intensity):
        self.intensity = intensity
        self.setPattern()
    
    def setPattern(self, ledpattern=None):
        # sends pattern with proper intensity
        if ledpattern is not None:
            self.ledpattern = ledpattern
        pattern2send = (self.ledpattern>=1)*self.intensity
        self.send_LEDMatrix_array(pattern2send, timeout=self.timeout)
        return self.ledpattern
    
    def getPattern(self):
        return self.ledpattern
        
    def set_led(self, colour=(0,0,0)):
        payload = {
            "red": colour[0],
            "green": colour[1],
            "blue": colour[2]
        }
        path = '/led'
        r = self._parent.post_json(path, payload)
        return r

    def setLEDArrayConfig(self, ledArrPin=4, ledArrNum=25):
        return self._parent.config.setLEDArrayConfig(ledArrPin=4, ledArrNum=25)
        